@@ -1,16 +1,26 @@
 # ext/declarative/extensions.py
-# Copyright (C) 2005-2022 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2025 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
 # the MIT License: https://www.opensource.org/licenses/mit-license.php
+# mypy: ignore-errors
+
+
 """Public API functions and helpers for declarative."""
+from __future__ import annotations
 
+import collections
+import contextlib
+from typing import Any
+from typing import Callable
+from typing import TYPE_CHECKING
+from typing import Union
 
-from ... import inspection
-from ... import util
+from ... import exc as sa_exc
+from ...engine import Connection
+from ...engine import Engine
 from ...orm import exc as orm_exc
-from ...orm import registry
 from ...orm import relationships
 from ...orm.base import _mapper_or_none
 from ...orm.clsregistry import _resolver
@@ -19,25 +29,11 @@ from ...orm.util import polymorphic_union
 from ...schema import Table
 from ...util import OrderedDict
 
-
-@util.deprecated(
-    "2.0",
-    "the instrument_declarative function is deprecated "
-    "and will be removed in SQLAlhcemy 2.0.  Please use "
-    ":meth:`_orm.registry.map_declaratively",
-)
-def instrument_declarative(cls, cls_registry, metadata):
-    """Given a class, configure the class declaratively,
-    using the given registry, which can be any dictionary, and
-    MetaData object.
-
-    """
-    registry(metadata=metadata, class_registry=cls_registry).map_declaratively(
-        cls
-    )
+if TYPE_CHECKING:
+    from ...sql.schema import MetaData
 
 
-class ConcreteBase(object):
+class ConcreteBase:
     """A helper class for 'concrete' declarative mappings.
 
     :class:`.ConcreteBase` will use the :func:`.polymorphic_union`
@@ -54,23 +50,26 @@ class ConcreteBase(object):
 
         from sqlalchemy.ext.declarative import ConcreteBase
 
+
         class Employee(ConcreteBase, Base):
-            __tablename__ = 'employee'
+            __tablename__ = "employee"
             employee_id = Column(Integer, primary_key=True)
             name = Column(String(50))
             __mapper_args__ = {
-                            'polymorphic_identity':'employee',
-                            'concrete':True}
+                "polymorphic_identity": "employee",
+                "concrete": True,
+            }
+
 
         class Manager(Employee):
-            __tablename__ = 'manager'
+            __tablename__ = "manager"
             employee_id = Column(Integer, primary_key=True)
             name = Column(String(50))
             manager_data = Column(String(40))
             __mapper_args__ = {
-                            'polymorphic_identity':'manager',
-                            'concrete':True}
-
+                "polymorphic_identity": "manager",
+                "concrete": True,
+            }
 
     The name of the discriminator column used by :func:`.polymorphic_union`
     defaults to the name ``type``.  To suit the use case of a mapping where an
@@ -79,7 +78,7 @@ class ConcreteBase(object):
     ``_concrete_discriminator_name`` attribute::
 
         class Employee(ConcreteBase, Base):
-            _concrete_discriminator_name = '_concrete_discriminator'
+            _concrete_discriminator_name = "_concrete_discriminator"
 
     .. versionadded:: 1.3.19 Added the ``_concrete_discriminator_name``
        attribute to :class:`_declarative.ConcreteBase` so that the
@@ -133,19 +132,23 @@ class AbstractConcreteBase(ConcreteBase):
     :class:`.AbstractConcreteBase` will use the :func:`.polymorphic_union`
     function automatically, against all tables mapped as a subclass
     to this class.   The function is called via the
-    ``__declare_last__()`` function, which is essentially
-    a hook for the :meth:`.after_configured` event.
+    ``__declare_first__()`` function, which is essentially
+    a hook for the :meth:`.before_configured` event.
 
-    :class:`.AbstractConcreteBase` does produce a mapped class
-    for the base class, however it is not persisted to any table; it
-    is instead mapped directly to the "polymorphic" selectable directly
-    and is only used for selecting.  Compare to :class:`.ConcreteBase`,
-    which does create a persisted table for the base class.
+    :class:`.AbstractConcreteBase` applies :class:`_orm.Mapper` for its
+    immediately inheriting class, as would occur for any other
+    declarative mapped class. However, the :class:`_orm.Mapper` is not
+    mapped to any particular :class:`.Table` object.  Instead, it's
+    mapped directly to the "polymorphic" selectable produced by
+    :func:`.polymorphic_union`, and performs no persistence operations on its
+    own.  Compare to :class:`.ConcreteBase`, which maps its
+    immediately inheriting class to an actual
+    :class:`.Table` that stores rows directly.
 
     .. note::
 
-        The :class:`.AbstractConcreteBase` class does not intend to set up  the
-        mapping for the base class until all the subclasses have been defined,
+        The :class:`.AbstractConcreteBase` delays the mapper creation of the
+        base class until all the subclasses have been defined,
         as it needs to create a mapping against a selectable that will include
         all subclass tables.  In order to achieve this, it waits for the
         **mapper configuration event** to occur, at which point it scans
@@ -155,84 +158,100 @@ class AbstractConcreteBase(ConcreteBase):
         While this event is normally invoked automatically, in the case of
         :class:`.AbstractConcreteBase`, it may be necessary to invoke it
         explicitly after **all** subclass mappings are defined, if the first
-        operation is to be a query against this base class.  To do so, invoke
-        :func:`.configure_mappers` once all the desired classes have been
-        configured::
-
-            from sqlalchemy.orm import configure_mappers
-
-            configure_mappers()
-
-        .. seealso::
-
-            :func:`_orm.configure_mappers`
+        operation is to be a query against this base class. To do so, once all
+        the desired classes have been configured, the
+        :meth:`_orm.registry.configure` method on the :class:`_orm.registry`
+        in use can be invoked, which is available in relation to a particular
+        declarative base class::
 
+            Base.registry.configure()
 
     Example::
 
+        from sqlalchemy.orm import DeclarativeBase
         from sqlalchemy.ext.declarative import AbstractConcreteBase
 
+
+        class Base(DeclarativeBase):
+            pass
+
+
         class Employee(AbstractConcreteBase, Base):
             pass
 
+
         class Manager(Employee):
-            __tablename__ = 'manager'
+            __tablename__ = "manager"
             employee_id = Column(Integer, primary_key=True)
             name = Column(String(50))
             manager_data = Column(String(40))
 
             __mapper_args__ = {
-                'polymorphic_identity':'manager',
-                'concrete':True}
+                "polymorphic_identity": "manager",
+                "concrete": True,
+            }
 
-        configure_mappers()
+
+        Base.registry.configure()
 
     The abstract base class is handled by declarative in a special way;
     at class configuration time, it behaves like a declarative mixin
     or an ``__abstract__`` base class.   Once classes are configured
     and mappings are produced, it then gets mapped itself, but
     after all of its descendants.  This is a very unique system of mapping
-    not found in any other SQLAlchemy system.
+    not found in any other SQLAlchemy API feature.
 
     Using this approach, we can specify columns and properties
     that will take place on mapped subclasses, in the way that
     we normally do as in :ref:`declarative_mixins`::
 
+        from sqlalchemy.ext.declarative import AbstractConcreteBase
+
+
         class Company(Base):
-            __tablename__ = 'company'
+            __tablename__ = "company"
             id = Column(Integer, primary_key=True)
 
+
         class Employee(AbstractConcreteBase, Base):
+            strict_attrs = True
+
             employee_id = Column(Integer, primary_key=True)
 
             @declared_attr
             def company_id(cls):
-                return Column(ForeignKey('company.id'))
+                return Column(ForeignKey("company.id"))
 
             @declared_attr
             def company(cls):
                 return relationship("Company")
 
+
         class Manager(Employee):
-            __tablename__ = 'manager'
+            __tablename__ = "manager"
 
             name = Column(String(50))
             manager_data = Column(String(40))
 
             __mapper_args__ = {
-                'polymorphic_identity':'manager',
-                'concrete':True}
+                "polymorphic_identity": "manager",
+                "concrete": True,
+            }
 
-        configure_mappers()
+
+        Base.registry.configure()
 
     When we make use of our mappings however, both ``Manager`` and
     ``Employee`` will have an independently usable ``.company`` attribute::
 
-        session.query(Employee).filter(Employee.company.has(id=5))
+        session.execute(select(Employee).filter(Employee.company.has(id=5)))
 
-    .. versionchanged:: 1.0.0 - The mechanics of :class:`.AbstractConcreteBase`
-       have been reworked to support relationships established directly
-       on the abstract base, without any special configurational steps.
+    :param strict_attrs: when specified on the base class, "strict" attribute
+     mode is enabled which attempts to limit ORM mapped attributes on the
+     base class to only those that are immediately present, while still
+     preserving "polymorphic" loading behavior.
+
+     .. versionadded:: 2.0
 
     .. seealso::
 
@@ -240,6 +259,8 @@ class AbstractConcreteBase(ConcreteBase):
 
         :ref:`concrete_inheritance`
 
+        :ref:`abstract_concrete_base`
+
     """
 
     __no_table__ = True
@@ -280,27 +301,46 @@ class AbstractConcreteBase(ConcreteBase):
         # In that case, ensure we update the properties entry
         # to the correct column from the pjoin target table.
         declared_cols = set(to_map.declared_columns)
+        declared_col_keys = {c.key for c in declared_cols}
         for k, v in list(to_map.properties.items()):
             if v in declared_cols:
                 to_map.properties[k] = pjoin.c[v.key]
+                declared_col_keys.remove(v.key)
 
         to_map.local_table = pjoin
 
+        strict_attrs = cls.__dict__.get("strict_attrs", False)
+
         m_args = to_map.mapper_args_fn or dict
 
         def mapper_args():
             args = m_args()
             args["polymorphic_on"] = pjoin.c[discriminator_name]
+            args["polymorphic_abstract"] = True
+            if strict_attrs:
+                args["include_properties"] = (
+                    set(pjoin.primary_key)
+                    | declared_col_keys
+                    | {discriminator_name}
+                )
+                args["with_polymorphic"] = ("*", pjoin)
             return args
 
         to_map.mapper_args_fn = mapper_args
 
-        m = to_map.map()
+        to_map.map()
 
-        for scls in cls.__subclasses__():
+        stack = [cls]
+        while stack:
+            scls = stack.pop(0)
+            stack.extend(scls.__subclasses__())
             sm = _mapper_or_none(scls)
-            if sm and sm.concrete and cls in scls.__bases__:
-                sm._set_concrete_base(m)
+            if sm and sm.concrete and sm.inherits is None:
+                for sup_ in scls.__mro__[1:]:
+                    sup_sm = _mapper_or_none(sup_)
+                    if sup_sm:
+                        sm._set_concrete_base(sup_sm)
+                        break
 
     @classmethod
     def _sa_raise_deferred_config(cls):
@@ -315,7 +355,7 @@ class AbstractConcreteBase(ConcreteBase):
         )
 
 
-class DeferredReflection(object):
+class DeferredReflection:
     """A helper class for construction of mappings based on
     a deferred reflection step.
 
@@ -335,10 +375,12 @@ class DeferredReflection(object):
 
         from sqlalchemy.ext.declarative import declarative_base
         from sqlalchemy.ext.declarative import DeferredReflection
+
         Base = declarative_base()
 
+
         class MyClass(DeferredReflection, Base):
-            __tablename__ = 'mytable'
+            __tablename__ = "mytable"
 
     Above, ``MyClass`` is not yet mapped.   After a series of
     classes have been defined in the above fashion, all tables
@@ -360,17 +402,22 @@ class DeferredReflection(object):
         class ReflectedOne(DeferredReflection, Base):
             __abstract__ = True
 
+
         class ReflectedTwo(DeferredReflection, Base):
             __abstract__ = True
 
+
         class MyClass(ReflectedOne):
-            __tablename__ = 'mytable'
+            __tablename__ = "mytable"
+
 
         class MyOtherClass(ReflectedOne):
-            __tablename__ = 'myothertable'
+            __tablename__ = "myothertable"
+
 
         class YetAnotherClass(ReflectedTwo):
-            __tablename__ = 'yetanothertable'
+            __tablename__ = "yetanothertable"
+
 
         # ... etc.
 
@@ -388,58 +435,123 @@ class DeferredReflection(object):
     """
 
     @classmethod
-    def prepare(cls, engine):
-        """Reflect all :class:`_schema.Table` objects for all current
-        :class:`.DeferredReflection` subclasses"""
+    def prepare(
+        cls, bind: Union[Engine, Connection], **reflect_kw: Any
+    ) -> None:
+        r"""Reflect all :class:`_schema.Table` objects for all current
+        :class:`.DeferredReflection` subclasses
+
+        :param bind: :class:`_engine.Engine` or :class:`_engine.Connection`
+         instance
+
+         ..versionchanged:: 2.0.16 a :class:`_engine.Connection` is also
+         accepted.
+
+        :param \**reflect_kw: additional keyword arguments passed to
+         :meth:`_schema.MetaData.reflect`, such as
+         :paramref:`_schema.MetaData.reflect.views`.
+
+         .. versionadded:: 2.0.16
+
+        """
 
         to_map = _DeferredMapperConfig.classes_for_base(cls)
 
-        with inspection.inspect(engine)._inspection_context() as insp:
+        metadata_to_table = collections.defaultdict(set)
+
+        # first collect the primary __table__ for each class into a
+        # collection of metadata/schemaname -> table names
+        for thingy in to_map:
+            if thingy.local_table is not None:
+                metadata_to_table[
+                    (thingy.local_table.metadata, thingy.local_table.schema)
+                ].add(thingy.local_table.name)
+
+        # then reflect all those tables into their metadatas
+
+        if isinstance(bind, Connection):
+            conn = bind
+            ctx = contextlib.nullcontext(enter_result=conn)
+        elif isinstance(bind, Engine):
+            ctx = bind.connect()
+        else:
+            raise sa_exc.ArgumentError(
+                f"Expected Engine or Connection, got {bind!r}"
+            )
+
+        with ctx as conn:
+            for (metadata, schema), table_names in metadata_to_table.items():
+                metadata.reflect(
+                    conn,
+                    only=table_names,
+                    schema=schema,
+                    extend_existing=True,
+                    autoload_replace=False,
+                    **reflect_kw,
+                )
+
+            metadata_to_table.clear()
+
+            # .map() each class, then go through relationships and look
+            # for secondary
             for thingy in to_map:
-                cls._sa_decl_prepare(thingy.local_table, insp)
                 thingy.map()
+
                 mapper = thingy.cls.__mapper__
                 metadata = mapper.class_.metadata
+
                 for rel in mapper._props.values():
                     if (
                         isinstance(rel, relationships.RelationshipProperty)
-                        and rel.secondary is not None
+                        and rel._init_args.secondary._is_populated()
                     ):
-                        if isinstance(rel.secondary, Table):
-                            cls._reflect_table(rel.secondary, insp)
-                        elif isinstance(rel.secondary, str):
+                        secondary_arg = rel._init_args.secondary
 
+                        if isinstance(secondary_arg.argument, Table):
+                            secondary_table = secondary_arg.argument
+                            metadata_to_table[
+                                (
+                                    secondary_table.metadata,
+                                    secondary_table.schema,
+                                )
+                            ].add(secondary_table.name)
+                        elif isinstance(secondary_arg.argument, str):
                             _, resolve_arg = _resolver(rel.parent.class_, rel)
 
-                            rel.secondary = resolve_arg(rel.secondary)
-                            rel.secondary._resolvers += (
-                                cls._sa_deferred_table_resolver(
-                                    insp, metadata
-                                ),
+                            resolver = resolve_arg(
+                                secondary_arg.argument, True
+                            )
+                            metadata_to_table[
+                                (metadata, thingy.local_table.schema)
+                            ].add(secondary_arg.argument)
+
+                            resolver._resolvers += (
+                                cls._sa_deferred_table_resolver(metadata),
                             )
 
-                            # controversy!  do we resolve it here? or leave
-                            # it deferred?   I think doing it here is necessary
-                            # so the connection does not leak.
-                            rel.secondary = rel.secondary()
+                            secondary_arg.argument = resolver()
+
+            for (metadata, schema), table_names in metadata_to_table.items():
+                metadata.reflect(
+                    conn,
+                    only=table_names,
+                    schema=schema,
+                    extend_existing=True,
+                    autoload_replace=False,
+                )
 
     @classmethod
-    def _sa_deferred_table_resolver(cls, inspector, metadata):
-        def _resolve(key):
-            t1 = Table(key, metadata)
-            cls._reflect_table(t1, inspector)
-            return t1
+    def _sa_deferred_table_resolver(
+        cls, metadata: MetaData
+    ) -> Callable[[str], Table]:
+        def _resolve(key: str) -> Table:
+            # reflection has already occurred so this Table would have
+            # its contents already
+            return Table(key, metadata)
 
         return _resolve
 
-    @classmethod
-    def _sa_decl_prepare(cls, local_table, inspector):
-        # autoload Table, which is already
-        # present in the metadata.  This
-        # will fill in db-loaded columns
-        # into the existing Table object.
-        if local_table is not None:
-            cls._reflect_table(local_table, inspector)
+    _sa_decl_prepare = True
 
     @classmethod
     def _sa_raise_deferred_config(cls):
@@ -450,14 +562,3 @@ class DeferredReflection(object):
             "method is called on the class hierarchy."
             % orm_exc._safe_cls_name(cls),
         )
-
-    @classmethod
-    def _reflect_table(cls, table, inspector):
-        Table(
-            table.name,
-            table.metadata,
-            extend_existing=True,
-            autoload_replace=False,
-            autoload_with=inspector,
-            schema=table.schema,
-        )
